import os
import re
import requests
import shutil
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple, List
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from feature_engine.encoding import RareLabelEncoder
from sklearn.preprocessing import OneHotEncoder

from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
from scipy import stats

# Calculate performance metrics
def calculate_metrics(y_true, y_pred):
    mse = mean_squared_error(y_true, y_pred)
    rmse = np.sqrt(mse)
    mae = mean_absolute_error(y_true, y_pred)
    r2 = r2_score(y_true, y_pred)
    
    return {
        'MSE': mse,
        'RMSE': rmse,
        'MAE': mae,
        'R²': r2
    }
    

def download_dataset(google_drive_link):
    # Extract file ID from Google Drive link
    match = re.search(r'/d/([a-zA-Z0-9_-]+)', google_drive_link)
    if not match:
        raise ValueError("Invalid Google Drive link.")
    file_id = match.group(1)
    
    # Create download URL
    download_url = f"https://drive.google.com/uc?export=download&id={file_id}"
    
    response = requests.get(download_url)
    if response.status_code != 200:
        raise Exception("Error downloading the file.")
    dataset_dir = "dataset"
    os.makedirs(dataset_dir, exist_ok=True)
    
    dest_path = os.path.join(dataset_dir, "dataset.csv")
    with open(dest_path, "wb") as f:
        f.write(response.content)

def DATAF_separate_columns(df: pd.DataFrame, output_col: str) -> Tuple[List[str], List[str]]:
    """Separate DataFrame columns into numerical and categorical lists."""
    cols = df.columns.tolist()
    input_cols = [col for col in cols if col != output_col]
    numerical_cols = df[input_cols].select_dtypes(include=['number']).columns.tolist()
    categorical_cols = [col for col in input_cols if col not in numerical_cols]
    return numerical_cols, categorical_cols

def DATAF_separate_num_columns(df: pd.DataFrame, num_cols: List[str], unique_threshold: int = 20) -> Tuple[List[str], List[str]]:
    """
    Separates a list of numerical columns into discrete and continuous lists based on the number of unique values.
    """
    discrete_cols = []
    continuous_cols = []
    
    for col in num_cols:
        # Count unique values. By default nunique() ignores NaNs.
        unique_count = df[col].nunique()
        if unique_count < unique_threshold:
            discrete_cols.append(col)
        else:
            continuous_cols.append(col)
    
    return discrete_cols, continuous_cols



def EDA_analyze_inputf(input_col: str, output_col: str, df: pd.DataFrame, unique_threshold: int = 20,
                       input_lambda=lambda x: x, target_lambda=lambda x: x) -> None:
    """
    Visualize the distribution and relation of an input feature with the output side by side,
    and print summary statistics including total samples, missing values, and unique values if applicable.
    
    If the input or target variable is numerical, apply the corresponding lambda transformation
    (by default, an identity lambda is used). This allows for custom transformation of the data
    before visualization and analysis.
    
    Parameters:
        input_col (str): Name of the input feature column.
        output_col (str): Name of the target variable column.
        df (pd.DataFrame): The DataFrame containing the data.
        unique_threshold (int, optional): Threshold to decide if a numeric variable is treated as discrete.
        input_lambda (function, optional): Transformation function for the input feature (default is identity).
        target_lambda (function, optional): Transformation function for the target variable (default is identity).
    """
    # Apply lambda transformations if the columns are numeric.
    if pd.api.types.is_numeric_dtype(df[input_col]):
        transformed_input = df[input_col].apply(input_lambda)
    else:
        transformed_input = df[input_col]
    
    if pd.api.types.is_numeric_dtype(df[output_col]):
        transformed_output = df[output_col].apply(target_lambda)
    else:
        transformed_output = df[output_col]
    
    df_temp = df.copy()
    df_temp[input_col] = transformed_input
    df_temp[output_col] = transformed_output

    fig, axes = plt.subplots(1, 2, figsize=(14, 5))
    
    # Determine variable type based on the input column.
    if pd.api.types.is_numeric_dtype(df_temp[input_col]) or df_temp[input_col].nunique() > unique_threshold:
        unique_count = df_temp[input_col].nunique()
        if unique_count < unique_threshold:
            variable_type = 'discrete'
            # Plot histogram for discrete numeric variable.
            sns.histplot(df_temp[input_col].dropna(), ax=axes[0])
            axes[0].set_title(f"Distribution of {input_col} (Discrete)")
            
            # Group by the input variable and plot median ± std of the target variable.
            grouped = df_temp.groupby(input_col)[output_col].agg(['median', 'std']).reset_index()
            axes[1].errorbar(grouped[input_col], grouped['median'], yerr=grouped['std'], fmt='-o')
            axes[1].set_title(f"{input_col} vs {output_col} (median ± std)")
        else:
            variable_type = 'continuous'
            # Plot histogram for continuous numeric variable.
            sns.histplot(df_temp[input_col].dropna(), ax=axes[0])
            axes[0].set_title(f"Distribution of {input_col} (Continuous)")
            
            # Group by the input variable and plot median ± std of the target variable.
            grouped = df_temp.groupby(input_col)[output_col].agg(['median', 'std']).reset_index()
            axes[1].errorbar(grouped[input_col], grouped['median'], yerr=grouped['std'], fmt='-o')
            axes[1].set_title(f"{input_col} vs {output_col} (median ± std)")
    else:
        variable_type = 'categorical'
        # Plot count plot for categorical variable.
        sns.countplot(x=input_col, data=df_temp, ax=axes[0])
        axes[0].set_title(f"Distribution of {input_col}")
        axes[0].tick_params(axis='x', rotation=45)
        
        # Plot boxplot for categorical vs target variable.
        sns.boxplot(x=input_col, y=output_col, data=df_temp, ax=axes[1])
        axes[1].set_title(f"{input_col} vs {output_col}")
        axes[1].tick_params(axis='x', rotation=45)

    plt.tight_layout()
    plt.show()

    # Print summary statistics.
    total_samples = len(df_temp[input_col])
    num_nans = df_temp[input_col].isna().sum()

    print("Total samples:", total_samples)
    print("Total missing values:", num_nans)
    print("Variable type:", variable_type)

    # If the variable is discrete or categorical, print its unique values.
    if variable_type in ['discrete', 'categorical']:
        unique_vals = df_temp[input_col].dropna().unique()
        print("Unique values:", unique_vals)

class ProbabilisticCategoricalImputer(BaseEstimator, TransformerMixin):
    """
    Imputes missing categorical values by randomly sampling from a given probability distribution.
    """
    def __init__(self, variables, probabilities):
        """
        Parameters:
        ----------
        variables : str or list
            The column(s) to impute
        probabilities : dict
            Dictionary with categories as keys and probabilities as values
        """
        self.variables = [variables] if isinstance(variables, str) else variables
        self.probabilities = probabilities
        
    def fit(self, X, y=None):
        return self
        
    def transform(self, X):
        X = X.copy()
        
        for variable in self.variables:
            nan_indices = X[variable].isna()
            n_nans = nan_indices.sum()
            
            if n_nans > 0:
                # Extract categories and their probabilities
                categories = list(self.probabilities.keys())
                probs = list(self.probabilities.values())
                
                random_values = np.random.choice(categories, size=n_nans, p=probs)
                
                X.loc[nan_indices, variable] = random_values
                
        return X

        

class LabelMapper(BaseEstimator, TransformerMixin):
    def __init__(self, mapping: dict):
        if not isinstance(mapping, dict):
            raise ValueError("The mapping should be a dictionary where keys are column names and values are mapping dictionaries.")
        self.mapping = mapping

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        X = X.copy()

        # Loop over each column to be mapped:
        for col, map_dict in self.mapping.items():
            if col not in X.columns:
                print(f"Error: Column '{col}' not found in the DataFrame.")
                continue  # Skip transformation for missing column.
            # Replace the column labels with numeric encoding.
            X[col] = X[col].map(map_dict)
        return X

class ColumnRemover(BaseEstimator, TransformerMixin):
    def __init__(self, variables_to_remove):
        self.variables_to_remove = variables_to_remove
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        X_copy = X.copy()
        # Filter columns to keep only those not in variables_to_remove
        columns_to_keep = [col for col in X_copy.columns if col not in self.variables_to_remove]
        return X_copy[columns_to_keep]
        

class SqrtTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, variables):
        self.variables = variables
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        X_copy = X.copy()
        for var in self.variables:
            if var in X_copy.columns:
                X_copy[var] = np.sqrt(X_copy[var])
        return X_copy



        
class RenameDropTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, rename_dict, drop_cols):
        self.rename_dict = rename_dict
        self.drop_cols = drop_cols
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        X_copy = X.copy()
        # Rename columns
        X_copy = X_copy.rename(columns=self.rename_dict)
        # Drop columns
        X_copy = X_copy.drop(columns=self.drop_cols, errors='ignore')
        return X_copy

        
class TemporalVariableTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, variables, reference_variable):
        
        if not isinstance(variables, list):
            raise ValueError('variables should be a list')
        
        self.variables = variables
        self.reference_variable = reference_variable

    def fit(self, X, y=None):
        return self

    def transform(self, X):

        X = X.copy()
        
        for feature in self.variables:
            X[feature] = X[self.reference_variable] - X[feature]

        return X


class OneHotEncodingTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, columns, sparse_output=False, handle_unknown='ignore'):
        """
        Parameters:
        - columns: list of column names to one-hot encode.
        - sparse_output: If True, returns a sparse matrix.
        - handle_unknown: How to handle unknown categories during transform.
        """
        self.columns = columns
        self.sparse_output = sparse_output
        self.handle_unknown = handle_unknown
        self.encoder = OneHotEncoder(sparse_output=self.sparse_output, 
                                     handle_unknown=self.handle_unknown)
        
    def fit(self, X, y=None):
        # Fit the encoder on the specified columns
        self.encoder.fit(X[self.columns])
        return self
        
    def transform(self, X):
        X_copy = X.copy()
        encoded_array = self.encoder.transform(X_copy[self.columns])
        encoded_df = pd.DataFrame(encoded_array, 
                                  columns=self.encoder.get_feature_names_out(self.columns),
                                  index=X_copy.index)
        # Drop the original categorical columns
        X_copy = X_copy.drop(columns=self.columns)
        # Concatenate the one-hot encoded DataFrame with the original data
        X_copy = pd.concat([X_copy, encoded_df], axis=1)
        return X_copy

